#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import time
import logging
import datetime

import sqlalchemy as sa

from sqlalchemy.ext.declarative         import declarative_base
from sqlalchemy                         import orm
from sqlalchemy                         import BigInteger, Boolean, Column, DateTime, Integer, String, ForeignKey
from sqlalchemy.orm                     import relationship
from sqlalchemy.ext.hybrid              import hybrid_property

from modules.base.config                import AuraConfig
from modules.base.utils                 import SimpleUtil
from modules.core.resources             import ResourceUtil



# Init Config
config = AuraConfig()

# Initialize DB Model and session
engine = sa.create_engine(config.get_database_uri())
Base = declarative_base()
Base.metadata.bind = engine

class DB():
    session = orm.scoped_session(orm.sessionmaker())(bind=engine)
    Model = Base



class AuraDatabaseModel():
    """
    AuraDataBaseModel.

    Holding all tables and relationships for the engine.
    """
    logger = None


    def __init__(self):
        """
        Constructor.
        """
        self.logger = logging.getLogger("AuraEngine")


    def store(self, add=False, commit=False):
        """
        Store to the database
        """
        if add:
            DB.session.add(self)
        else:
            DB.session.merge(self)
        if commit:
            DB.session.commit()


    def delete(self, commit=False):
        """
        Delete from the database
        """
        DB.session.delete(self)
        if commit:
            DB.session.commit()


    def refresh(self):
        """
        Refreshes the currect record
        """
        DB.session.expire(self)
        DB.session.refresh(self)


    def _asdict(self):
        return self.__dict__


    @staticmethod
    def recreate_db(systemexit = False):
        """
        Re-creates the database for developments purposes.
        """
        manualschedule = Schedule()
        manualschedule.schedule_id = 0
        manualschedule.show_name = "Manual Show"
        Base.metadata.drop_all()
        Base.metadata.create_all()

        
        # self.logger.debug("inserting manual scheduling possibility and fallback trackservice schedule")
        # DB.session.add(manualschedule)
        # db.session.add(fallback_trackservice_schedule)
        # self.logger.debug("all created. commiting...")
        DB.session.commit()

        if systemexit:
            sys.exit(0)



#
#   SCHEDULES & PLAYLISTS
#


class Schedule(DB.Model, AuraDatabaseModel):
    """
    One specific Schedule for a show on a timeslot.
    Holding references to playlists and fallback-playlists.
    """
    __tablename__ = 'schedule'

    # Primary keys
    id = Column(Integer, primary_key=True, autoincrement=True)

    schedule_start = Column(DateTime, unique=True, index=True)
    schedule_end = Column(DateTime, unique=True, index=True)
    schedule_id = Column(Integer, unique=True)

    show_id = Column(Integer)
    show_name = Column(String(256))
    show_hosts = Column(String(256))
    funding_category = Column(String(256))
    comment = Column(String(512))
    languages = Column(String(256))
    type = Column(String(256))
    category = Column(String(256))
    topic = Column(String(256))
    musicfocus = Column(String(256))

    is_repetition = Column(Boolean())

    playlist_id = Column(Integer) #, ForeignKey("playlist.playlist_id"))
    schedule_fallback_id = Column(Integer)
    show_fallback_id = Column(Integer)
    station_fallback_id = Column(Integer)
    
    fadeouttimer = None # Used to fade-out the schedule, even when entries are longer

    playlist = relationship("Playlist",
                            primaryjoin="and_(Schedule.schedule_start==Playlist.schedule_start, Schedule.playlist_id==Playlist.playlist_id, Schedule.show_name==Playlist.show_name)",
                            back_populates="schedule")
    schedule_fallback = relationship("Playlist",
                            primaryjoin="and_(Schedule.schedule_start==Playlist.schedule_start, Schedule.schedule_fallback_id==Playlist.playlist_id, Schedule.show_name==Playlist.show_name)",
                            back_populates="schedule")
    show_fallback = relationship("Playlist",
                            primaryjoin="and_(Schedule.schedule_start==Playlist.schedule_start, Schedule.show_fallback_id==Playlist.playlist_id, Schedule.show_name==Playlist.show_name)",
                            back_populates="schedule")
    station_fallback = relationship("Playlist",
                            primaryjoin="and_(Schedule.schedule_start==Playlist.schedule_start, Schedule.station_fallback_id==Playlist.playlist_id, Schedule.show_name==Playlist.show_name)",
                            back_populates="schedule")


    @staticmethod
    def select_show_on_datetime(date_time):
        return DB.session.query(Schedule).filter(Schedule.schedule_start == date_time).first()


    @staticmethod
    def select_programme(date_from=datetime.date.today()):
        """
        Select all schedules starting from `date_from` or from today if no
        parameter is passed.

        Args:
            date_from (datetime):   Select schedules from this date and time on

        Returns:
            ([Schedule]):           List of schedules
        """
        schedules = DB.session.query(Schedule).\
            filter(Schedule.schedule_start >= date_from).\
            order_by(Schedule.schedule_start).all()

        return schedules


    @staticmethod
    def select_upcoming(n):
        """
        Selects the (`n`) upcoming schedules.
        """
        now = datetime.datetime.now()
        DB.session.commit() # Required since independend session is used.
        schedules = DB.session.query(Schedule).\
            filter(Schedule.schedule_start > str(now)).\
            order_by(Schedule.schedule_start.asc()).limit(n).all()
        
        return schedules


    def get_playlist(self):
        """
        Returns the assigned playlist.
        """
        # TODO Refactor to avoid storing array of playlists.
        if self.playlist and self.playlist[0]:
            return self.playlist[0]
        return None


    def has_queued_entries(self):
        """
        Checks if entries of this timeslot have been queued at the engine.        
        """
        #TODO Make logic more transparent
        if hasattr(self, "queued_entries"):
            if len(self.queued_entries) > 0:    
                return True
        return False


    @hybrid_property
    def start_unix(self):
        """
        Start time of the schedule in UNIX time.
        """
        return time.mktime(self.schedule_start.timetuple())


    @hybrid_property
    def end_unix(self):
        """
        End time of the schedule in UNIX time.
        """
        return time.mktime(self.schedule_end.timetuple())


    def as_dict(self):
        """
        Returns the schedule as a dictionary for serialization.
        """
        playlist = self.playlist

        return {
            "schedule_id": self.schedule_id,        
            "schedule_start": self.schedule_start.isoformat(),
            "schedule_end": self.schedule_end.isoformat(),

            "topic": self.topic,
            "musicfocus": self.musicfocus,
            "funding_category": self.funding_category,
            "is_repetition": self.is_repetition,
            "category": self.category,
            "languages": self.languages,
            "comment": self.comment,
            "playlist_id": self.playlist_id,
            "schedule_fallback_id": self.schedule_fallback_id,
            "show_fallback_id": self.show_fallback_id,
            "station_fallback_id": self.station_fallback_id,

            "show": {
                "name": self.show_name,
                "host": self.show_hosts
            },

            "playlist": playlist
        }


    def __str__(self):
        """
        String representation of the object.
        """
        time_start = SimpleUtil.fmt_time(self.start_unix)
        time_end = SimpleUtil.fmt_time(self.end_unix)
        return "ID#%s [Show: %s, ShowID: %s | %s - %s ]" % (str(self.schedule_id), self.show_name, str(self.show_id), time_start, time_end)



class Playlist(DB.Model, AuraDatabaseModel):
    """
    The playlist containing playlist entries.
    """
    __tablename__ = 'playlist'

    # pk,fk
    artificial_id = Column(Integer, primary_key=True)
    schedule_start = Column(DateTime, ForeignKey("schedule.schedule_start"))
    # relationships
    schedule = relationship("Schedule", uselist=False, back_populates="playlist")
    entries = relationship("PlaylistEntry", back_populates="playlist")
    # data
    playlist_id = Column(Integer, autoincrement=False)  # , ForeignKey("schedule.playlist_id"))
    show_name = Column(String(256))
    entry_count = Column(Integer)


    @staticmethod
    def select_all():
        """
        Fetches all entries
        """
        all_entries = DB.session.query(Playlist).filter(Playlist.fallback_type == 0).all()

        cnt = 0
        for entry in all_entries:
            entry.programme_index = cnt
            cnt = cnt + 1

        return all_entries


    @staticmethod
    def select_playlist_for_schedule(start_date, playlist_id):
        """
        Retrieves the playlist for the given schedule identified by `start_date` and `playlist_id`

        Args:
            start_date (datetime):  Date and time when the playlist is scheduled
            playlist_id (Integer):  The ID of the playlist

        Returns:
            (Playlist):             The playlist, if existing for schedule

        Raises:
            Exception:              In case there a inconsistent database state, such es multiple playlists for given date/time.
        """
        playlist = None
        playlists = DB.session.query(Playlist).filter(Playlist.schedule_start == start_date).all()
        # FIXME There are unknown issues with the native SQL query by ID
        # playlists = DB.session.query(Playlist).filter(Playlist.schedule_start == datetime and Playlist.playlist_id == playlist_id).all()
        
        for p in playlists:
            if p.playlist_id == playlist_id:
                playlist = p

        return playlist


    @staticmethod
    def select_playlist(playlist_id):
        """
        Retrieves all paylists for that given playlist ID.

        Args:
            playlist_id (Integer):  The ID of the playlist
        
        Returns:
            (Array<Playlist>):      An array holding the playlists
        """
        return DB.session.query(Playlist).filter(Playlist.playlist_id == playlist_id).order_by(Playlist.schedule_start).all()
    

    @staticmethod
    def is_empty():
        """
        Checks if the given is empty
        """
        try:
            return not DB.session.query(Playlist).one_or_none()
        except sa.orm.exc.MultipleResultsFound:
            return False


    @hybrid_property
    def start_unix(self):
        """
        Start time of the playlist in UNIX time.
        """
        return time.mktime(self.schedule_start.timetuple())


    @hybrid_property
    def end_unix(self):
        """
        End time of the playlist in UNIX time.
        """
        return time.mktime(self.schedule_start.timetuple()) + self.duration


    @hybrid_property
    def duration(self):
        """
        Returns the total length of the playlist in seconds.

        Returns:
            (Integer):  Length in seconds
        """
        total = 0

        for entry in self.entries:
            total += entry.duration
        return total


    def as_dict(self):
        """
        Returns the playlist as a dictionary for serialization.
        """
        entries = []
        for e in self.entries:
            entries.append(e.as_dict())

        playlist = {
            "playlist_id": self.playlist_id,
            "fallback_type": self.fallback_type,
            "entry_count": self.entry_count,
            "entries": entries
        }
        return playlist


    def __str__(self):
        """
        String representation of the object.
        """
        time_start = SimpleUtil.fmt_time(self.start_unix)
        time_end = SimpleUtil.fmt_time(self.end_unix)
        return "ID#%s [items: %s | %s - %s]" % (str(self.playlist_id), str(self.entry_count), str(time_start), str(time_end))



class PlaylistEntry(DB.Model, AuraDatabaseModel):
    """
    Playlist entries are the individual items of a playlist such as audio files.
    """
    __tablename__ = 'playlist_entry'

    # primary keys
    artificial_id = Column(Integer, primary_key=True)

    # foreign keys
    artificial_playlist_id = Column(Integer, ForeignKey("playlist.artificial_id"))
    entry_num = Column(Integer) # , primary_key=True)

    uri = Column(String(1024))
    duration = Column(BigInteger)
    source = Column(String(1024))
    entry_start = Column(DateTime)
    entry_start_actual = None # Assigned when the entry is actually played
    channel = None # Assigned when entry is actually played
    queue_state = None # Assigned when entry is about to be queued    
    status = None # Assigned when state changes
    
    switchtimer = None
    loadtimer = None
    fadeouttimer = None

    # relationships
    playlist = relationship("Playlist", uselist=False, back_populates="entries")
    meta_data = relationship("PlaylistEntryMetaData", uselist=False, back_populates="entry")

    @staticmethod
    def select_playlistentry_for_playlist(artificial_playlist_id, entry_num):
        """
        Selects one entry identified by `playlist_id` and `entry_num`.
        """
        return DB.session.query(PlaylistEntry).filter(PlaylistEntry.artificial_playlist_id == artificial_playlist_id, PlaylistEntry.entry_num == entry_num).first()

    @staticmethod
    def delete_entry(artificial_playlist_id, entry_num):
        """
        Deletes the playlist entry and associated metadata.
        """
        entry = PlaylistEntry.select_playlistentry_for_playlist(artificial_playlist_id, entry_num)
        metadata = PlaylistEntryMetaData.select_metadata_for_entry(entry.artificial_id)
        metadata.delete()
        entry.delete()
        DB.session.commit()

    @staticmethod
    def count_entries(artificial_playlist_id):
        """
        Returns the count of all entries.
        """
        result = DB.session.query(PlaylistEntry).filter(PlaylistEntry.artificial_playlist_id == artificial_playlist_id).count()
        return result

    @hybrid_property
    def entry_end(self):
        return self.entry_start + datetime.timedelta(seconds=self.duration)

    @hybrid_property
    def start_unix(self):
        return time.mktime(self.entry_start.timetuple())

    @hybrid_property
    def end_unix(self):
        return time.mktime(self.entry_end.timetuple())

    @hybrid_property
    def volume(self):
        return 100 # FIXME Make DB Column

    def get_content_type(self):
        return ResourceUtil.get_content_type(self.uri)


    def get_prev_entries(self):
        """
        Retrieves all previous entries as part of the current entry's playlist.

        Returns:
            (List):     List of PlaylistEntry
        """
        prev_entries = []
        for entry in self.playlist.entries:
            if entry.entry_start < self.entry_start:
                prev_entries.append(entry)
        return prev_entries


    def get_next_entries(self, schedule_sensitive=True):
        """
        Retrieves all following entries as part of the current entry's playlist.

        Args:
            schedule_sensitive (Boolean):   If `True` entries which start after \
                the end of the schedule are excluded

        Returns:
            (List):     List of PlaylistEntry
        """
        next_entries = []
        for entry in self.playlist.entries:
            if entry.entry_start > self.entry_start:
                if schedule_sensitive:
                    if entry.entry_start < self.playlist.schedule.schedule_end:
                        next_entries.append(entry)
                else:
                    next_entries.append(entry)
        return next_entries


    def as_dict(self):
        """
        Returns the entry as a dictionary for serialization.
        """
        if self.meta_data:
            return {
                "id": self.artificial_id,
                "duration": self.duration,
                "artist": self.meta_data.artist,
                "album": self.meta_data.album,
                "title": self.meta_data.title
            }
        return None

    def __str__(self):
        """
        String representation of the object.
        """
        time_start = SimpleUtil.fmt_time(self.start_unix)
        time_end = SimpleUtil.fmt_time(self.end_unix)
        track = self.source[-25:]
        return "PlaylistEntry #%s [%s - %s | %ssec | Source: ...%s]" % (str(self.artificial_id), time_start, time_end, self.duration, track)




class PlaylistEntryMetaData(DB.Model, AuraDatabaseModel):
    """
    Metadata for a playlist entry such as the artist and track name.
    """
    __tablename__ = "playlist_entry_metadata"

    artificial_id = Column(Integer, primary_key=True)
    artificial_entry_id = Column(Integer, ForeignKey("playlist_entry.artificial_id"))

    artist = Column(String(256))
    title = Column(String(256))
    album = Column(String(256))

    entry = relationship("PlaylistEntry", uselist=False, back_populates="meta_data")

    @staticmethod
    def select_metadata_for_entry(artificial_playlistentry_id):
        return DB.session.query(PlaylistEntryMetaData).filter(PlaylistEntryMetaData.artificial_entry_id == artificial_playlistentry_id).first()


