#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import datetime

from threading                  import Thread

from modules.base.config import AuraConfig
from modules.base.utils         import SimpleUtil as SU
from modules.base.mail          import AuraMailer
from modules.plugins.monitor    import AuraMonitor


from modules.plugins.trackservice import TrackServiceHandler


class EventBinding():
    """
    A binding between the event dispatcher and some event handler.

    It allows you to subscribe to events in a chained way:

        ```
        binding = dispatcher.attach(AuraMonitor)
        binding.subscribe("on_boot").subscribe("on_play")
        ```
    """
    dispatcher = None
    instance = None

    def __init__(self, dispatcher, instance):
        self.dispatcher = dispatcher
        self.instance = instance


    def subscribe(self, event_type):
        """
        Subscribes the instance to some event identified by the `event_type` string.
        """
        self.dispatcher.subscribe(self.instance, event_type)
        return self


    def get_instance(self):
        """
        Returns the object within that binding.
        """
        return self.instance



class EngineEventDispatcher():
    """
    Executes handlers for engine events.
    """
    logger = None
    config = None

    subscriber_registry = None
    mailer = None
    engine = None
    scheduler = None
    monitor = None


    def __init__(self, engine, scheduler):
        """
        Initialize EventDispatcher
        """
        self.subscriber_registry = dict()
        self.logger = logging.getLogger("AuraEngine")
        self.config = AuraConfig.config()
        self.mailer = AuraMailer(self.config)
        self.engine = engine
        self.scheduler = scheduler
        
        binding = self.attach(AuraMonitor)
        binding.subscribe("on_boot")
        binding.subscribe("on_sick")
        binding.subscribe("on_resurrect")

        binding = self.attach(TrackServiceHandler)
        binding.subscribe("on_timeslot")
        binding.subscribe("on_play")
        binding.subscribe("on_metadata")
        binding.subscribe("on_queue")


    def attach(self, clazz):
        """
        Creates an intance of the given Class.
        """
        instance = clazz(self.engine)
        return EventBinding(self, instance)


    def subscribe(self, instance, event_type):
        """
        Subscribes to some event type. Preferably use it via `EventBinding.subscribe(..)`.
        """
        if not event_type in self.subscriber_registry:
            self.subscriber_registry[event_type] = []
        self.subscriber_registry[event_type].append(instance)


    def call_event(self, event_type, args):
        """
        Calls all subscribers for the given event type.
        """
        if not event_type in self.subscriber_registry:
            return
        listeners = self.subscriber_registry[event_type]
        if not listeners:
            return
        for listener in listeners:
            method = getattr(listener, event_type)
            if method:
                if args:
                    method(args)
                else:
                    method()


    #
    # Events
    #


    def on_initialized(self):
        """
        Called when the engine is initialized e.g. connected to Liquidsoap.
        """
        self.logger.debug("on_initialized(..)")
        self.scheduler.on_initialized()
        self.call_event("on_initialized", None)


    def on_boot(self):
        """
        Called when the engine is starting up. This happens after the initialization step.
        """
        self.logger.debug("on_boot(..)")
        self.call_event("on_boot", None)


    def on_ready(self):
        """
        Called when the engine is booted and ready to play.
        """
        self.logger.debug("on_ready(..)")
        self.scheduler.on_ready()


    def on_timeslot(self, timeslot):
        """
        Event Handler which is called by the scheduler when the current timeslot is refreshed.

        Args:
            source (String):    The `PlaylistEntry` object
        """
        def func(self, timeslot):        
            self.logger.debug("on_timeslot(..)")
            self.call_event("on_timeslot", timeslot)

        thread = Thread(target = func, args = (self, timeslot))
        thread.start()    


    def on_play(self, entry):
        """
        Event Handler which is called by the engine when some entry is actually playing. 

        Args:
            source (String):    The `PlaylistEntry` object
        """
        def func(self, entry):        
            self.logger.debug("on_play(..)")
            # Assign timestamp for play time
            entry.entry_start_actual = datetime.datetime.now()
            self.call_event("on_play", entry)

        thread = Thread(target = func, args = (self, entry))
        thread.start()    


    def on_metadata(self, data):
        """
        Event Handler which is called by the soundsystem implementation (i.e. Liquidsoap)
        when some entry is actually playing.

        Args:
            data (dict):    A collection of metadata related to the current track
        """
        def func(self, data):        
            self.logger.debug("on_metadata(..)")
            self.call_event("on_metadata", data)

        thread = Thread(target = func, args = (self, data))
        thread.start() 


    def on_stop(self, entry):
        """
        The entry on the assigned channel has been stopped playing.
        """
        def func(self, entry):        
            self.logger.debug("on_stop(..)")
            self.call_event("on_stop", entry)
        
        thread = Thread(target = func, args = (self, entry))
        thread.start() 


    def on_fallback_updated(self, playlist_uri):
        """
        Called when the scheduled fallback playlist has been updated.
        """
        self.logger.debug("on_fallback_updated(..)")
        self.call_event("on_fallback_updated", playlist_uri)


    def on_fallback_cleaned(self, cleaned_channel):
        """
        Called when the scheduled fallback queue has been cleaned up.
        """
        self.logger.debug("on_fallback_cleaned(..)")
        self.call_event("on_fallback_cleaned", cleaned_channel)


    def on_idle(self):
        """
        Callend when no entry is playing
        """
        def func(self):
            self.logger.debug("on_idle(..)")
            self.logger.error(SU.red("Currently there's nothing playing!"))
            self.call_event("on_idle", None)

        thread = Thread(target = func, args = (self, ))
        thread.start() 


    def on_schedule_change(self, schedule):
        """
        Called when the playlist or entries of the current schedule have changed.
        """
        def func(self, schedule):        
            self.logger.debug("on_schedule_change(..)")
            self.call_event("on_schedule_change", schedule)

        thread = Thread(target = func, args = (self, schedule))
        thread.start() 


    def on_queue(self, entries):
        """
        One or more entries have been queued and are currently pre-loaded.
        """
        def func(self, entries):        
            self.logger.debug("on_queue(..)")
            self.call_event("on_queue", entries)

        thread = Thread(target = func, args = (self, entries))
        thread.start() 


    def on_sick(self, data):
        """
        Called when the engine is in some unhealthy state.
        """
        def func(self, data):        
            self.logger.debug("on_sick(..)")
            self.call_event("on_sick", data)

        thread = Thread(target = func, args = (self, data))
        thread.start() 


    def on_resurrect(self, data):
        """
        Called when the engine turned healthy again after being sick.
        """
        def func(self, data):        
            self.logger.debug("on_resurrect(..)")
            self.call_event("on_resurrect", data)

        thread = Thread(target = func, args = (self, data))
        thread.start() 


    def on_critical(self, subject, message, data=None):
        """
        Callend when some critical event occurs
        """
        def func(self, subject, message, data):        
            self.logger.debug("on_critical(..)")
            if not data: data = ""
            self.mailer.send_admin_mail(subject, message + "\n\n" + str(data))
            self.call_event("on_critical", (subject, message, data))

        thread = Thread(target = func, args = (self, subject, message, data))
        thread.start() 