
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging
import requests

from datetime import datetime
from modules.base.utils import SimpleUtil as SU


class CalendarFetcher:
    """
    Fetches the schedules, playlists and playlist entries as JSON
    via the API endpoints of Steering and Tank.
    """
    config = None
    logging = None
    has_already_fetched = False
    fetched_schedule_data = None

    # Config for API Endpoints
    steering_calendar_url = None
    tank_playlist_url = None
    tank_session = None
    tank_secret = None



    def __init__(self, config):
        """
        Constructor

        Args:
            config (AuraConfig):    Holds the engine configuration
        """
        self.config = config
        self.logger = logging.getLogger("AuraEngine")
        self.steering_calendar_url = self.config.get("api_steering_calendar")
        self.tank_playlist_url = self.config.get("api_tank_playlist")
        self.tank_session = self.config.get("api_tank_session")
        self.tank_secret = self.config.get("api_tank_secret")


    #
    #   PUBLIC METHODS
    #


    def fetch(self):
        """
        Retrieve all required data from the API.
        """
        return_data = []

        self.logger.debug("Fetching schedules from STEERING")
        self.fetched_schedule_data = self.fetch_schedule_data()
        if not self.fetched_schedule_data:
            self.logger.critical(SU.red("No schedules fetched from API!"))
            return None

        self.logger.debug("Fetching playlists from TANK")
        self.fetch_playlists()

        try:
            for schedule in self.fetched_schedule_data:
                # Skip schedule if no start or end is given
                if "start" not in schedule:
                    self.logger.warning("No start of schedule given. Skipping schedule: " + str(schedule))
                    schedule = None
                if "end" not in schedule:
                    self.logger.warning("No end of schedule given. Skipping schedule: " + str(schedule))
                    schedule = None
                if "playlist" not in schedule \
                    and "show_fallback" not in schedule \
                    and "schedule_fallback" not in schedule \
                    and "station_fallback" not in schedule:
                    
                    self.logger.warning("No playlist for schedule given. Skipping schedule: " + str(schedule))
                    schedule = None

                if schedule:
                    return_data.append(schedule)
        except TypeError:
            self.logger.error(SU.red("Nothing fetched ..."))
            self.fetched_schedule_data = None
            return None

        return return_data



    #
    #   PRIVATE METHODS
    #


    def fetch_schedule_data(self):
        """
        Fetches schedule data from Steering.

        Returns:
            ([Schedule]):   An array of schedules
        """
        schedule = None
        headers = { "content-type": "application/json" }
       
        try:
            self.logger.debug("Fetch schedules from Steering API...")                
            response = requests.get(self.steering_calendar_url, data=None, headers=headers)
            if not response.status_code == 200:
                self.logger.critical(SU.red("HTTP Status: %s | Schedules could not be fetched! Response: %s" % \
                    (str(response.status_code), response.text)))
                return None            
            schedule = response.json()                
        except Exception as e:
            self.logger.critical(SU.red("Error while requesting schedules from Steering!"), e)

        if not schedule:
            self.logger.error(SU.red("Got no schedule via Playout API (Steering)!"))
            return None

        return self.remove_unnecessary_data(schedule)



    def fetch_playlists(self):
        """
        Fetches all playlists including fallback playlists for every schedule.
        This method used the class member `fetched_schedule_data`` to iterate
        over and extend schedule data.
        """
        # store fetched entries => do not have to fetch playlist_id more than once
        fetched_entries=[]
        local_station_fallback_id = str(self.config.get("scheduling_station_fallback_id"))

        try:
            for schedule in self.fetched_schedule_data:

                # Get IDs of playlists
                playlist_id = self.get_playlist_id(schedule, "playlist_id")
                schedule_fallback_id = self.get_playlist_id(schedule, "schedule_fallback_id")
                show_fallback_id = self.get_playlist_id(schedule, "show_fallback_id")
                station_fallback_id = self.get_playlist_id(schedule, "station_fallback_id")

                # Retrieve playlist and the fallback playlists for every schedule.
                # If a playlist (like station_fallback) is already fetched, it is not fetched again but reused
                schedule["playlist"]          = self.fetch_playlist(playlist_id,          fetched_entries)
                schedule["schedule_fallback"] = self.fetch_playlist(schedule_fallback_id, fetched_entries)
                schedule["show_fallback"]     = self.fetch_playlist(show_fallback_id,     fetched_entries)
                schedule["station_fallback"]  = self.fetch_playlist(station_fallback_id,  fetched_entries)

                # If Steering doesn't provide a station fallback, the local one is used
                if not schedule["station_fallback"] and int(local_station_fallback_id) > 0:
                    schedule["station_fallback_id"] = local_station_fallback_id 
                    schedule["station_fallback"] = self.fetch_playlist(local_station_fallback_id, fetched_entries)
                    if schedule["station_fallback"]:
                        self.logger.info("Assigned playlist #%s as local station fallback to schedule #%s" % (local_station_fallback_id, schedule["schedule_id"]))

        except Exception as e:
            self.logger.error("Error while fetching playlists from API endpoints: " + str(e), e)



    def fetch_playlist(self, playlist_id, fetched_playlists):
        """
        Fetches the playlist for a given schedule.

        Args:
            id_name (String):       The type of playlist to fetch (e.g. normal vs. fallback)
            fetched_playlists ([]): Previously fetched playlists to avoid re-fetching

        Returns:
            (Playlist):             Playlist of type `id_name`
        """
        if not playlist_id:
            return None
        playlist = None
        url = self.tank_playlist_url.replace("${ID}", playlist_id) 
        headers = {
            "Authorization": "Bearer %s:%s" % (self.tank_session, self.tank_secret), 
            "content-type": "application/json"
        }

        # If playlist is already fetched in this round, use the existing one
        for playlist in fetched_playlists:
            if playlist["id"] == playlist_id:
                self.logger.debug("Playlist #%s already fetched" % playlist_id)
                return playlist
                   
        try:
            self.logger.debug("Fetch playlist from Tank API...")             
            response = requests.get(url, data=None, headers=headers)
            if not response.status_code == 200:
                self.logger.critical(SU.red("HTTP Status: %s | Playlist #%s could not be fetched or is not available! Response: %s" % \
                    (str(response.status_code), str(playlist_id), response.text)))
                return None          
            playlist = response.json()                  
        except Exception as e:
            self.logger.critical(SU.red("Error while requesting playlist #%s from Tank" % str(playlist_id)), e)
            return None

        fetched_playlists.append(playlist)
        return playlist
 


    def get_playlist_id(self, schedule, id_name):
        """
        Extracts the playlist ID for a given playlist (fallback) type.
        """
        playlist_id = str(schedule[id_name])
        if not playlist_id or playlist_id == "None":
            self.logger.debug("No value defined for '%s' in schedule '#%s'" % (id_name, schedule["schedule_id"]))
            return None
        
        return playlist_id


    def remove_unnecessary_data(self, schedule):
        """
        Removes all schedules which are not relevant for 
        further processing.
        """
        count_before = len(schedule)
        schedule = self.remove_data_more_than_24h_in_the_future(schedule)
        schedule = self.remove_data_in_the_past(schedule)
        count_after = len(schedule)

        self.logger.debug("Removed %d unnecessary schedules from response. Entries left: %d" % ((count_before - count_after), count_after))
        return schedule


    def remove_data_more_than_24h_in_the_future(self, schedules):
        """ 
        Removes entries 24h in the future and 12 hours in the past.
        Note: This might influence resuming (in case of a crash)  
        single schedules which are longer than 12 hours long.
        Think e.g. live broadcasts.
        """
        items = []
        now = SU.timestamp()
        now_plus_24hours = now + (12*60*60)
        now_minus_12hours = now - (12*60*60)

        for s in schedules:
            start_time = datetime.strptime(s["start"], "%Y-%m-%dT%H:%M:%S")
            start_time = SU.timestamp(start_time)

            if start_time <= now_plus_24hours and start_time >= now_minus_12hours:
                items.append(s)

        return items


    def remove_data_in_the_past(self, schedules):
        """
        Removes all schedules from the past, except the one which is 
        currently playing.
        """
        items = []
        now = SU.timestamp()
        for s in schedules:
            start_time = datetime.strptime(s["start"], "%Y-%m-%dT%H:%M:%S")
            start_time = SU.timestamp(start_time)
            end_time = datetime.strptime(s["end"], "%Y-%m-%dT%H:%M:%S")
            end_time = SU.timestamp(end_time)

            # Append all elements in the future
            if start_time >= now:
                items.append(s)
             # Append the one which is playing now
            elif start_time < now < end_time:
                items.append(s)

        return items
