
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging

from enum                       import Enum

from modules.base.config        import AuraConfig
from modules.base.utils         import SimpleUtil as SU
from modules.base.mail          import AuraMailer
from modules.core.resources     import ResourceClass
from modules.core.channels      import Channel
from modules.core.control       import EngineExecutor



class FallbackType(Enum):
    """
    Types of playlists.
    """
    NONE        = { "id": 0, "name": "default", "lqs_sources": [ Channel.QUEUE_A, Channel.QUEUE_A] }    # No fallback active, default playout
    SCHEDULE    = { "id": 1, "name": "schedule", "lqs_sources": [ Channel.FALLBACK_QUEUE_A, Channel.FALLBACK_QUEUE_B]}   # The first played when some default playlist fails
    SHOW        = { "id": 2, "name": "show", "lqs_sources": [ "station_folder", "station_playlist"]}       # The second played when the timeslot fallback fails    
    STATION     = { "id": 3, "name": "station", "lqs_sources": [ "station_folder", "station_playlist"] }    # The last played when everything else fails

    @property
    def id(self):
        return self.value["id"]

    def __str__(self):
        return str(self.value["name"])



class FallbackManager:
    """
    Handles all types of fallbacks in case there is an outage
    for the regular radio programme.

    Attributes:
        config (AuraConfig):        The engine configuration
        logger (AuraLogger):        The logger
        mail   (AuraMailer):        Mail service
        scheduler (AuraScheduler):  The scheduler
    """    
    config = None
    logger = None
    mailer = None
    scheduler = None

    
    def __init__(self, scheduler):
        """
        Constructor

        Args:

        """
        self.config = AuraConfig.config()
        self.logger = logging.getLogger("AuraEngine")
        self.mailer = AuraMailer(self.config)
        self.scheduler = scheduler


    #
    #   METHODS
    #


    def queue_fallback_playlist(self, timeslot):
        """
        Evaluates the scheduled fallback and queues it using a timed thread.
        """
        (fallback_type, playlist) = self.get_fallback_playlist(timeslot)

        if playlist:
            self.logger.info(f"Resolved {fallback_type.value} fallback")         
            return FallbackCommand(timeslot, playlist.entries)
        else:
            msg = f"There is no schedule- or show-fallback defined for timeslot#{timeslot.schedule_id}. "
            msg += f"The station fallback will be used automatically."
            self.logger.info(msg)



    def resolve_playlist(self, timeslot):
        """
        Retrieves the currently planned (fallback) playlist. If a normal playlist is available,
        this one is returned. In case of station fallback no playlist is returned.

        Args:
            timeslot (Schedule)
        
        Returns:
            (FallbackType, Playlist)
        """
        planned_playlist = None
        fallback_type = None

        if self.validate_playlist(timeslot, "playlist"):
            planned_playlist = timeslot.get_playlist()
            fallback_type = FallbackType.NONE
        else:
            (fallback_type, planned_playlist) = self.get_fallback_playlist(timeslot)            

        return (fallback_type, planned_playlist)



    def get_fallback_playlist(self, timeslot):
        """
        Retrieves the playlist to be used in a fallback scenario.

        Args: 
            timeslot (Schedule)

        Returns:
            (Playlist)
        """        
        playlist = None
        fallback_type = FallbackType.STATION

        if self.validate_playlist(timeslot, "schedule_fallback"):
            playlist = timeslot.schedule_fallback[0]
            fallback_type = FallbackType.SCHEDULE
        elif self.validate_playlist(timeslot, "show_fallback"):
            playlist = timeslot.show_fallback[0]
            fallback_type = FallbackType.SHOW

        return (fallback_type, playlist)



    def validate_playlist(self, timeslot, playlist_type):
        """
        Checks if a playlist is valid for play-out.

        Following checks are done for all playlists:

            - has one or more entries

        Fallback playlists should either:
                
            - have filesystem entries only
            - reference a recording of a previous playout of a show (also filesystem)
        
        Otherwise, if a fallback playlist contains Live or Stream entries,
        the exact playout behaviour can hardly be predicted.
        """
        playlist = getattr(timeslot, playlist_type)
        if playlist \
            and isinstance(playlist, list) \
            and playlist[0].entries \
            and len(playlist[0].entries) > 0:

            # Default playlist
            if playlist_type == "playlist":
                return True

            # Fallback playlist
            elif playlist[0].entries:
                is_fs_only = True
                for entry in playlist[0].entries:
                    if entry.get_content_type() not in ResourceClass.FILE.types:
                        self.logger.error(SU.red("Fallback playlist of type '%s' contains not only file-system entries! \
                            Skipping fallback level..." % playlist_type))
                        is_fs_only = False
                        break                
                return is_fs_only

        return False






class FallbackCommand(EngineExecutor):
    """
    Command composition for executing timed scheduling and unscheduling of fallback playlists.

    Based on the `timeslot.start_date` and `timeslot.end_date` two `EngineExecutor commands
    are created.
    """

    def __init__(self, timeslot, entries):
        """
        Constructor

        Args:
            timeslot (Timeslot):    The timeslot any fallback entries should be scheduled for
            entries (List):         List of entries to be scheduled as fallback
        """        
        from modules.core.engine import Engine

        def do_play(entries):
            self.logger.info(SU.cyan(f"=== start_fallback_playlist('{entries}') ==="))
            Engine.get_instance().player.start_fallback_playlist(entries)

        def do_stop():
            self.logger.info(SU.cyan("=== stop_fallback_playlist() ==="))
            Engine.get_instance().player.stop_fallback_playlist()

        child_timer = EngineExecutor("FALLBACK", None, timeslot.end_unix, do_stop, None)
        super().__init__("FALLBACK", child_timer, timeslot.start_unix, do_play, entries)